require 'set'
require_relative 'utils/interval'

class MergeIntervals
  # Definition for an interval.
  # class Interval
  #     attr_accessor :start, :end
  #     def initialize(s=0, e=0)
  #         @start = s
  #         @end = e
  #     end
  # end

  # @param {Interval[]} intervals
  # @return {Interval[]}
  # Sorting
  # def merge(intervals)
  #   merged = []
  #   intervals.sort_by! { |interval| interval.start }
  #
  #   intervals.each do |interval|
  #     if merged.empty? || merged.last.end < interval.start
  #       merged << interval
  #     else
  #       merged.last.end = interval.end if interval.end > merged.last.end
  #     end
  #   end
  #
  #   return merged
  # end

  # Connected Components
  def merge(intervals)
    graph = build_graph(intervals)
    nodes_in_comp, comp_num =get_components(graph, intervals)

    (0...comp_num).to_a.map { |comp| merge_nodes(nodes_in_comp[comp])}
  end

  private

  def overlap(node1, node2)
    return node1.start <= node2.end && node2.start <= node1.end
  end

  def build_graph(intervals)
    graph = Hash.new []

    intervals.each.with_index do |interval, i|
      for j in (i + 1)...intervals.length
        if overlap(interval, intervals[j])
          graph[interval] << intervals[j]
          graph[intervals[j]] << interval
        end
      end
    end

    return graph
  end

  def merge_nodes(nodes)
    min_start = nodes.min_by { |node| node.start }
    max_end = nodes.max_by { |node| node.end }
    Interval.new(min_start, max_end)
  end

  def get_components(graph, intervals)
    visited = Set.new
    comp_num = 0
    nodes_in_comp = Hash.new []

    make_component_dfs = Proc.new do |start|
      stack = [start]
      while stack
        node = stack.pop()
        unless visited.include?(node)
          visited.add(node)
          nodes_in_comp[comp_num] << node
          stack.concat(graph[node])
        end
      end
    end

    intervals.each do |interval|
      unless visited.include?(interval)
        make_component_dfs.call(interval)
        return nodes_in_comp, comp_num
      end
    end
  end
end